#!/usr/bin/env python3
import logging
import logging.config
import os

import yaml


def setup_logging(
    path='logging.yaml',
    default_level=logging.INFO
):
    """Setup logging configuration

    """
    _create_trace_loglevel(logging)
    if os.path.exists(path):
        with open(path, 'rt') as f:
            config = yaml.safe_load(f.read())
        logging.config.dictConfig(config)
    else:
        logging.basicConfig(level=default_level)

def get(name):
    logger = logging.getLogger(name)
    return logger

def _create_trace_loglevel(logging):
    "Add TRACE log level and Logger.trace() method."

    logging.TRACE = 5
    logging.addLevelName(logging.TRACE, "TRACE")

    def _trace(logger, message, *args, **kwargs):
        if logger.isEnabledFor(logging.TRACE):
            logger._log(logging.TRACE, message, args, **kwargs)

    logging.Logger.trace = _trace



#!/usr/bin/env python3
from dataclasses import dataclass, field
from typing import List

from xmltv.models import Channel


@dataclass(frozen=True)
class ChannelKey:
    id: str = field(compare=True, hash=True)
    channel: Channel = field(compare=False, hash=False)



#!/usr/bin/env python3

def clean_text(elem):
    text = ''
    for e in elem.descendants:
        if isinstance(e, str):
            text += e.strip()
        elif e.name == 'br':
            text += '\n'
    return text

def argparse_str2bool(v):
    if isinstance(v, bool):
        return v
    if v.lower() in ('yes', 'true', 't', 'y', '1'):
        return True
    elif v.lower() in ('no', 'false', 'f', 'n', '0'):
        return False
    else:
        raise argparse.ArgumentTypeError('Boolean value expected.')



#!/usr/bin/env python3
"""EPG Scraper for a specific website"""

from __future__ import annotations

import logging
from abc import ABC, abstractmethod
from datetime import date
from typing import Dict, List

from fake_useragent import UserAgent
from py_epg.common.requests import get_http_session
from xmltv.models import Channel, Programme

UA = UserAgent()


class EpgScraper(ABC):
    """Abstract class providing simple methods to fetch XMLTV data from an EPG website."""

    def __init__(self, name: str, proxy=None, user_agent=None):
        super().__init__()
        self._log = logging.getLogger(name)
        self._user_agent = user_agent if user_agent else UA.random
        self._http = get_http_session(user_agent=self._user_agent, proxy=proxy)

    @abstractmethod
    def site_name(self) -> str:
        """Returns the site_id of the EPG website this scraper supports"""

    @abstractmethod
    def fetch_channel(self, site_id, xmltv_id, name) -> Channel:
        """Fetches and returns the requested channel object."""

    @abstractmethod
    def fetch_programs(self, channel: Channel, channel_site_id: str, fetch_date: date) -> List[Programme]:
        """
        Returns a list of all programs for the given channel and day.

        Notes:
            - All dates must be in local xmltv format with timezone info.
            - Stop times are automatically set from programs' start times.
            - The order of returned programs is irrelevant, they are sorted by channel & start time.

        Parameters:
            channel_site_id: the channel's name as present on this EPG site
            date: the day of which programs need to be fetched for

        Returns:
            List[Programme]: the fetched channel and its programmes for the given 'day'.
        """



#!/usr/bin/env python3
"""Module defines request related configurations."""
from urllib3 import util

import requests
from requests import adapters


def get_http_session(
        retries=5,
        backoff_factor=20.0,
        status_forcelist=(500, 502, 504),
        session=None,
        proxy=None,
        user_agent=None,
) -> requests.Session:
    """
    Build request retry policy.
    wait bydefault to 5+ mins in 5 retries unless these settings overriden by client.
    """
    session = session or requests.Session()
    retry = util.Retry(
        total=retries,
        read=retries,
        connect=retries,
        backoff_factor=backoff_factor,
        status_forcelist=status_forcelist,
    )
    adapter = adapters.HTTPAdapter(max_retries=retry)
    session.mount('http://', adapter)
    session.mount('https://', adapter)
    if user_agent:
        session.headers.update({'User-Agent': user_agent})
    if proxy:
        session.proxies.update({'http': proxy, 'https': proxy})
    return session



#!/usr/bin/env python3

import argparse
import logging
import os
import pathlib
import sys
import time
import tqdm
from collections import defaultdict
from datetime import date, timedelta
from multiprocessing import Pool, current_process
from pprint import pprint
from typing import Dict, List, Set, Tuple

import requests
import xmltv
from dateutil.parser import parse
from lxml import etree as ET
from xmltv import xmltv_helpers
from xmltv.models import Channel, Programme, Tv

from py_epg.common.epg_scraper import EpgScraper
from py_epg.common.multiprocess_helper import setup_ltree_pickling
from py_epg.common.types import ChannelKey
from py_epg.common.utils import argparse_str2bool
from py_epg.scrapers import *

LOG = logging.getLogger(__name__)
DEFAULT_POOL_SIZE = 1
PBAR_NAME_COL_WIDTH = 15


class PyEPG:
    """Main class of PyEPG"""

    def __init__(self):
        self._log = logging.getLogger(__name__)
        self._args = self._parse_args()
        self._config = self._read_config()
        self._epg_scrapers = self._init_epg_scrapers()
        setup_ltree_pickling()
        pool_size = self._config.find('pool-size')
        self._pool = Pool(int(pool_size.text)
                          if pool_size is not None else DEFAULT_POOL_SIZE)

    def run(self):
        data = self._fetch_data()
        tv = self._build_xmltv(data)
        self._write_xmltv(tv)

    def _build_xmltv(self, data: Dict[ChannelKey, List[Programme]]):
        channels = []
        programs = []
        for chan_key, prgs in sorted(data.items(),
                                     key=lambda item: item[0].id):
            channels.append(chan_key.channel)
            programs.extend(
                sorted(prgs, key=lambda prg: prg and prg.start))
        self._post_process_programs(programs)
        return Tv(channels, programs, date=date.today().strftime('%Y%m%d%H%M%S'), generator_info_name='py_epg')

    def _post_process_programs(self, programs: List[Programme]):
        for i, program in enumerate(programs):
            # Set stop times
            if i < len(programs) - 1 and programs[i + 1].channel == program.channel:
                program.stop = programs[i + 1].start
            else:
                program.stop = f'{program.start[:8]}235959{program.start[14:]}'

    def _write_xmltv(self, tv: Tv):
        xmltv_out_file = pathlib.Path(self._config.find('filename').text)
        self._log.info(f'Writing results to {xmltv_out_file}..')
        xmltv_helpers.write_file_from_xml(xmltv_out_file, tv)

    def _fetch_data(self) -> Dict[ChannelKey, List[Programme]]:
        pbar_id = 'All Channels'
        programs_by_channel = defaultdict(list)
        channels = self._config.findall('channel')
        bar_unit_format = 'Progs: 0'
        bar_format = "{l_bar} {bar}|Chan: {n_fmt:>3}/{total_fmt:<3} [T:{elapsed} ETA:{remaining:<5}]"
        channel_programs = tqdm.tqdm(self._pool.imap_unordered(self._fetch_channel, channels, chunksize=1),
                                     total=len(channels),
                                     disable=not self._args.progress_bar,
                                     position=0,
                                     unit=bar_unit_format.format(0),
                                     dynamic_ncols=True,
                                     colour='cyan',
                                     bar_format=bar_format,
                                     postfix={'T': len(programs_by_channel)},
                                     desc=f'{pbar_id: >{PBAR_NAME_COL_WIDTH}}')
        for (chan_key, chan_progs) in channel_programs:
            programs_by_channel[chan_key].extend(chan_progs)
            prog_count = sum([len(listElem)
                             for listElem in programs_by_channel.values()])
            channel_programs.set_postfix({'T': prog_count})
            channel_programs.unit = bar_unit_format.format(
                len(programs_by_channel))

        return programs_by_channel

    def _fetch_channel(self, chan) -> Tuple[ChannelKey, List[Programme]]:
        site = chan.attrib['site']
        chan_site_id = chan.attrib['site_id']
        chan_xmltv_id = chan.attrib['xmltv_id']
        chan_name = chan.text

        scraper = self._epg_scrapers.get(site)
        if not scraper:
            self._log.error(f'Could not find scraper for site={site}.')
            sys.exit(-1)

        channel = scraper.fetch_channel(chan_site_id, chan_name)
        today = date.today()
        days = int(self._config.find('timespan').text)

        programs = []

        process = current_process()
        pbar_id = chan_site_id if len(chan_site_id) <= PBAR_NAME_COL_WIDTH - 2 \
            else chan_site_id[:PBAR_NAME_COL_WIDTH - 2] + '..'
        bar_format = "{l_bar} {bar}|Days: {n_fmt:>3}/{total_fmt:<3} [T:{elapsed} ETA:{remaining:<5}]"
        days_range = tqdm.tqdm(iterable=range(days),
                               disable=not self._args.progress_bar,
                               position=process._identity[0],
                               unit='day',
                               leave=False,
                               dynamic_ncols=True,
                               bar_format=bar_format,
                               colour='green',
                               desc=f'{pbar_id: >{PBAR_NAME_COL_WIDTH}}')
        for i in days_range:
            fetch_date = today + timedelta(days=i)
            day_programs = scraper.fetch_programs(
                channel, chan_site_id, fetch_date)
            key = ChannelKey(channel.id, channel)
            programs.extend(day_programs)
        return key, programs

    def _init_epg_scrapers(self) -> Dict[str, EpgScraper]:
        result = {}
        implementations = EpgScraper.__subclasses__()
        proxy = self._config.find('proxy')
        user_agent = self._config.find('user-agent')
        for scraper_class in implementations:
            obj = scraper_class(proxy=proxy.text if proxy is not None else None,
                                user_agent=user_agent.text if user_agent is not None else None)
            result[obj.site_name()] = obj
        return result

    def _read_config(self) -> Dict:
        return ET.parse(self._args.config)

    def _parse_args(self):
        # Initialize parser
        parser = argparse.ArgumentParser(
            prog='py_epg',
            description='A simple, multi-threaded, modular EPG grabber written in Python')
        parser.add_argument(
            "-p", "--progress-bar", help="Show a progress bar. Default: True",
            default=True, type=argparse_str2bool, nargs='?', const=True)
        parser.add_argument(
            "-q", "--quiet", help="Quiet mode (no progress-bar, etc). Default: False",
            default=False, type=argparse_str2bool, nargs='?', const=True)
        requiredArgs = parser.add_argument_group('required arguments')
        requiredArgs.add_argument(
            "-c", "--config", help="Path to py_epg.xml file", required=True)
        args = parser.parse_args()

        if args.quiet:
            args.progress_bar = False
        return args

    def __getstate__(self):
        self_dict = self.__dict__.copy()
        del self_dict['_pool']
        return self_dict

    def __setstate__(self, state):
        self.__dict__.update(state)


def main(args=None):
    py_epg = PyEPG()
    py_epg.run()



#!/usr/bin/env python3
from os import listdir
from os.path import basename, dirname

__all__ = [basename(f)[:-3] for f in listdir(dirname(__file__))
           if f[-3:] == ".py" and not f.endswith("__init__.py")]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['py_epg', 'py_epg.common', 'py_epg.scrapers']

package_data = \
{'': ['*']}

install_requires = \
['PyYAML>=6.0,<7.0',
 'beautifulsoup4>=4.10.0,<5.0.0',
 'fake-useragent>=0.1.11,<0.2.0',
 'lxml>=4.6.3,<5.0.0',
 'py-xmltv>=1.0.8,<2.0.0',
 'python-dateutil>=2.8.2,<3.0.0',
 'python-xmltv>=1.4.3,<2.0.0',
 'requests>=2.26.0,<3.0.0',
 'roman>=3.3,<4.0',
 'tqdm>=4.62.3,<5.0.0']

entry_points = \
{'console_scripts': ['epg = py_epg.main:main']}

setup_kwargs = {
    'name': 'py-epg',
    'version': '0.2.6',
    'description': 'py_epg is an easy to use, modular, multi-process EPG grabber written in Python.',
    'long_description': '# py-epg\n\n**py-epg** is an easy to use, modular, multi-process EPG grabber written in Python.\n\n* 📺 Scrapes various TV Program websites and saves programs in XMLTV format.\n* 🧩 Simply extend [EpgScraper](https://github.com/szab100/py_epg/blob/main/py_epg/common/epg_scraper.py) to grab EPG from your favorite TV site (requires basic Python skills).\n* 🤖 The framework provides the rest:\n    * [Beautiful Soup](https://www.crummy.com/software/BeautifulSoup/bs4/doc) - easily search & extract data from html elements \n    * multi-processing\n    * config management\n    * logging\n    * build & write XMLTV (with auto-generated fields, eg \'stop\')\n    * proxy server support\n    * auto http/s retries\n    * random fake user_agents\n* 🚀 Save time by fetching channels in parallel (caution: use proxy server(s) to avoid getting blacklisted)!\n* 🧑🏻\u200d💻 Your contributions are welcome! Feel free to create a PR with your tv-site scraper and/or framework improvements.\n\n<p align="center">\n  <img src="https://raw.githubusercontent.com/szab100/py_epg/main/py_epg.gif">\n</p>\n\n## Usage\n\n1. Install package:\n    ```sh\n    $ pip3 install py_epg\n    ```\n2. Create configuration: py_epg.xml\n    - Add all your channels (see [sample config](https://github.com/szab100/py-epg/blob/main/py_epg.xml)).\n    - Make sure there is a corresponding site scraper implementation in [py_epg/scrapers](https://github.com/szab100/py-epg/tree/main/py_epg/scrapers) for each channels (\'site\' attribute).\n3. Run:\n    ```sh\n    $ python3 -m py_epg -c </path/to/your/py_epg.xml>\n    ```\n\n    ..or see all supported flags:\n    ```sh\n    $ python3 -m py_epg -h\n    usage: py_epg [-h] [-p [PROGRESS_BAR]] [-q [QUIET]] -c CONFIG\n    ...\n    ```\n\n## Development\n\nYour contributions are welcome! Setup your dev environment as described below. [VSCode](https://code.visualstudio.com/) is a great free IDE for python projects. Once you are ready with your cool tv site scraper or framework feature, feel free to open a Pull Request here.\n\n1. Install poetry: \n    ```sh\n    curl -sSL https://raw.githubusercontent.com/python-poetry/poetry/master/get-poetry.py | python -\n    ```\n\n2. Clone repository & install dependencies:\n      ```sh\n      git clone https://github.com/szab100/py-epg.git && cd py-epg\n      ```\n\n3. Configure py_epg.xml\n    - Add all your channels (see the sample config xml). Make sure you have a scraper implementation in py_epg/scrapers/ for each channels (\'site\' attribute).\n\n4. Run:\n      ```sh\n      poetry install\n      poetry run epg -c py_epg.xml\n      ```\n\n## License\n\nCopyright 2021. Released under the MIT license.\n',
    'author': 'Szabolcs Fruhwald',
    'author_email': 'mail@szab100.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/szab100/py-epg',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



#!/usr/bin/env python3
import logging
import re
from datetime import date
from pprint import pprint
from string import Template
from typing import List

import roman
from bs4 import BeautifulSoup
from dateutil import tz
from dateutil.parser import parse
from py_epg.common.epg_scraper import EpgScraper
from py_epg.common.utils import clean_text
from xmltv.models import (Actor, Channel, Credits, Desc, DisplayName,
                          EpisodeNum, Icon, Programme, SubTitle, Title)

RE_MIXED_DESCRIPTION = re.compile(
    # Group 1 (opt): (title in orig lang)
    r'^(?:\((.*)\)\n\n)'
    # Group 2 (opt): description
    r'?(?:(.*?)(?=Rendezte:|Rendező:|Főszereplők:|$))?'
    # Group 3 (opt): director
    r'(?:(?:Rendezte:|Rendező:)\s*(.*?)(?=Főszereplők:|$))?'
    # Group 4 (opt): cast
    r'(?:\s*Főszereplők:\s*(.*))?',
    flags=re.S
)

RE_SEASON_EPISODE = re.compile(
    "((?=[MDCLXVI])M*D?C{0,4}L?X{0,4}V?I{0,4})\.\/([0-9]+)\.")
RE_EPISODE_RANGE = re.compile("([0-9]+)\.-([0-9]+)\.")
RE_SINGLE_EPISODE = re.compile("([0-9]+)\.")


class MusorTvMobile(EpgScraper):
    def __init__(self, proxy=None, user_agent=None):
        super().__init__(name=__name__, proxy=proxy, user_agent=user_agent)
        self._site_id = "m.musor.tv"
        self._base_url = 'https://m.musor.tv'
        self._page_encoding = 'utf-8'
        self._chan_id_tpl = Template('$chan_id.' + self._site_id)
        self._day_url_tpl = Template(
            self._base_url + '/napi/tvmusor/$chan_site_id/$date')
        self._tz_utc = tz.tzutc()
        self._tz_local = tz.tzlocal()

    def site_name(self) -> str:
        return self._site_id

    def fetch_channel(self, chan_site_id, name) -> Channel:
        today_str = date.today().strftime("%Y.%m.%d")
        url = self._day_url_tpl.substitute(
            chan_site_id=chan_site_id, date=today_str)
        soup = self._get_soup(url)
        channel_id = self._chan_id_tpl.substitute(chan_id=chan_site_id).upper()
        channel_logo = soup.select_one('img.channelheaderlink')
        channel_logo_src = self._base_url + \
            channel_logo.attrs['src'] if channel_logo else None
        return Channel(
            id=channel_id,
            display_name=[DisplayName(content=[name])],
            icon=Icon(src=channel_logo_src) if channel_logo_src else None)

    def fetch_programs(self, channel: Channel, channel_site_id: str, fetch_date: date) -> List[Programme]:
        date_str = fetch_date.strftime("%Y.%m.%d")
        channel_daily_progs_page = self._get_soup(self._day_url_tpl.substitute(
            chan_site_id=channel_site_id, date=date_str))
        programs_selector = 'section[itemscope]'
        programs = []
        for prg in channel_daily_progs_page.select(programs_selector):
            program = self._get_program(channel_site_id, fetch_date, prg)
            if program:
                programs.append(program)
        return programs

    def _get_program(self, chan_site_id: str, fetch_date: date, prg: BeautifulSoup) -> Programme:
        # 1. Fetch basic program info from the daily listing page
        channel_id = self._chan_id_tpl.substitute(chan_id=chan_site_id).upper()
        prg_title = prg.select_one('[itemprop="name"]').get_text(strip=True)
        program = Programme(channel=channel_id,
                            title=[Title(content=[prg_title], lang='hu')],
                            clumpidx=None)

        self._set_prg_sub_title_and_year(program, prg)
        self._set_prg_episode_info(program, prg_title)
        prg_start = self._set_prg_start(program, prg)

        # skip program starting < 00:00 or > 23:59
        if prg_start.date() != fetch_date:
            return None

        # 2. Fetch extended program info from program details page
        prg_details_link = prg.select_one(
            'h3.wideprogentry_progtitle > a').attrs['href']
        prg_details_page = self._get_soup(self._base_url + prg_details_link)

        self._set_prg_icon(program, prg_details_page)
        self._set_prg_fields_from_mixed_description(program, prg_details_page)

        self._log.trace(
            f'New program CH: {channel_id} ENC: {prg.original_encoding} P: {prg_title}')
        return program

    def _set_prg_start(self, program, prg):
        start = prg.select_one(
            'span[itemprop="startDate"]').attrs['content']
        start = parse(start.replace('GMT', 'T'))
        start = start.replace(tzinfo=self._tz_utc)
        start = start.astimezone(self._tz_local)
        program.start = start.strftime('%Y%m%d%H%M%S %z')
        return start

    def _set_prg_episode_info(self, program, title):
        # TV Shows - Season, Episode info in title
        m0 = RE_SEASON_EPISODE.search(title)
        # m1 = re_episode_range.search(title)
        m2 = RE_SINGLE_EPISODE.search(title)
        if m0 or m2:
            season = 0
            episode = 0
            if m0:
                season = roman.fromRoman(m0.group(1))
                episode = int(m0.group(2))
                title = title.split(str(m0.group()))[0].strip()
            if m2:
                episode = int(m2.group(1))
                title = title.split(str(m2.group()))[0].strip()
            onscreen = f'S{season:02d}E{episode:02d}' if season > 0 else f'S--E{episode:02d}'
            xmltv_ns = f'{season - 1}.{episode - 1}.' if season > 0 else f'.{episode - 1}.'
            program.episode_num = [EpisodeNum(content=[onscreen], system='onscreen'),
                                   EpisodeNum(content=[xmltv_ns], system='xmltv_ns')]

    def _set_prg_sub_title_and_year(self, program, prg):
        prg_sub_title = prg.select_one('div[itemprop="description"]')
        if prg_sub_title:
            subtitle = prg_sub_title.get_text(strip=True)
            parts = subtitle.split(',')
            if len(parts) > 1:
                year = parts[-1]
                if '-' in year:
                    # 2005-2010 => pick end year, eg. 2010
                    year = year.split('-')[-1]
                program.date = [year]
                subtitle = SubTitle(content=[','.join(parts[:-1])], lang='hu')
                program.sub_title = [subtitle] + program.sub_title
            else:
                # edge case: no subtitle, just a year
                if subtitle.isnumeric():
                    program.date = [subtitle]
                else:
                    program.sub_title = [subtitle]

    def _set_prg_icon(self, program, prg_details_page):
        prg_icon = prg_details_page.select_one('img[itemprop="image"]')
        if prg_icon:
            program.icon = [Icon(src=self._base_url + prg_icon.attrs['src'])]

    def _set_prg_fields_from_mixed_description(self, program, prg_details_page):
        prg_mixed_desc = clean_text(
            prg_details_page.select_one('div.eventinfolongdescinner')).strip()

        if prg_mixed_desc:
            if len(prg_mixed_desc.splitlines()) == 1:
                program.desc.append(Desc(content=[prg_mixed_desc]))
                return

            # pprint(prg_mixed_desc)
            result = RE_MIXED_DESCRIPTION.search(prg_mixed_desc)
            if result and len(result.groups()):
                # pprint(result.groups())
                # title in orig lang
                if result.group(1):
                    program.title.append(
                        Title(content=[result.group(1)], lang='en'))

                if result.group(2):
                    content = result.group(2).strip()
                    parts = content.split('\n\n')
                    if len(parts) >= 2:
                        # sub-title + description
                        program.sub_title.append(
                            SubTitle(content=[parts[0].strip()]))
                        desc = '\n'.join(parts[1:]).strip()
                        if desc:
                            program.desc.append(Desc(content=[desc]))
                    elif content:
                        # description only
                        program.desc.append(Desc(content=[content]))
                # director, cast
                if result.group(3) or result.group(4):
                    separator = re.compile('[,;]+ ')
                    credits = Credits()
                    if result.group(3):
                        credits.director = separator.split(
                            result.group(3).strip())
                    if result.group(4):
                        actors = separator.split(result.group(4).strip())
                        credits.actor = [Actor(content=[actor])
                                         for actor in actors]
                    program.credits = credits

    def _get_soup(self, url) -> BeautifulSoup:
        page = self._http.get(url)
        return BeautifulSoup(page.text, "html.parser")



#!/usr/bin/env python3

import copyreg as copy_reg
from io import BytesIO

from lxml import etree


def element_unpickler(data):
    return etree.fromstring(data)


def element_pickler(element):
    data = etree.tostring(element)
    return element_unpickler, (data,)


def elementtree_unpickler(data):
    data = BytesIO(data)
    return etree.parse(data)


def elementtree_pickler(tree):
    data = BytesIO()
    tree.write(data)
    return elementtree_unpickler, (data.getvalue(),)


def setup_ltree_pickling():
    copy_reg.pickle(etree._Element, element_pickler, element_unpickler)
    copy_reg.pickle(etree._ElementTree, elementtree_pickler,
                    elementtree_unpickler)
